import asyncio
import json
from dataclasses import dataclass
from typing import Dict

from graphql import ExecutionResult, GraphQLError, GraphQLSchema, format_error, parse, subscribe
from starlette import status
from starlette.types import Receive, Scope, Send
from starlette.websockets import Message, WebSocket

from gql.subscribe import PROTOCOL, MessageType, OperationMessage


@dataclass
class ConnectionContext:
    socket: WebSocket
    operations: Dict[str, asyncio.Future]


class Subscription:
    schema: GraphQLSchema

    def __init__(self, schema: GraphQLSchema) -> None:
        self.schema = schema

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        socket = WebSocket(scope, receive=receive, send=send)
        await socket.accept(PROTOCOL)

        context = ConnectionContext(socket=socket, operations={})
        await self.on_message(context)

    async def on_message(self, context: ConnectionContext) -> None:
        close_code = status.WS_1000_NORMAL_CLOSURE
        try:
            while True:
                message = await context.socket.receive()
                if message["type"] == "websocket.receive":
                    await self.dispatch(context, message)
                elif message["type"] == "websocket.disconnect":
                    close_code = int(message.get("code", status.WS_1000_NORMAL_CLOSURE))
                    break
        except Exception as exc:
            close_code = status.WS_1011_INTERNAL_ERROR
            raise exc from None
        finally:
            await context.socket.close(close_code)

    async def dispatch(self, context: ConnectionContext, data: Message) -> None:
        message = await self.decode(context, data)
        if message.type == MessageType.GQL_CONNECTION_INIT:
            await self.init(context)
        elif message.type == MessageType.GQL_CONNECTION_TERMINATE:
            await self.terminate(context)
        elif message.type == MessageType.GQL_START:
            future = asyncio.ensure_future(self.start(context, message))
            context.operations[message.id] = future
        elif message.type == MessageType.GQL_STOP:
            await self.stop(context, message.id)
        else:
            for op in context.operations.values():
                op.cancel()
            await context.socket.close(code=status.WS_1003_UNSUPPORTED_DATA)

    async def init(self, context: ConnectionContext) -> None:
        await self.send_message(context, MessageType.GQL_CONNECTION_ACK)

    async def stop(self, context: ConnectionContext, op_id: str) -> None:
        if op_id in context.operations:
            op = context.operations[op_id]
            op.cancel()
            self.unsubscribe(context, op_id)
        await self.complete(context, op_id)

    async def terminate(self, context: ConnectionContext) -> None:
        for op in context.operations.values():
            op.cancel()
        await context.socket.close(code=status.WS_1000_NORMAL_CLOSURE)

    async def complete(self, context: ConnectionContext, op_id: str) -> None:
        await self.send_message(context, MessageType.GQL_COMPLETE, op_id=op_id)

    async def start(self, context: ConnectionContext, message: OperationMessage) -> None:
        # if message.id in self.operations:
        #     await self.unsubscribe(op_id=message.id)
        payload = message.payload
        assert payload
        try:
            doc = parse(payload.query)
        except GraphQLError as error:
            await self.send_execution_result(message.id, ExecutionResult(data=None, errors=[error]))
            return

        result_or_iterator = await subscribe(
            self.schema,
            doc,
            variable_values=payload.variables,
            operation_name=payload.operation_name,
        )
        if isinstance(result_or_iterator, ExecutionResult):
            await self.send_execution_result(context, message.id, result_or_iterator)
            return

        async for result in result_or_iterator:
            await self.send_execution_result(context, message.id, result)

        await self.complete(context, message.id)

    def unsubscribe(self, context: ConnectionContext, op_id: str) -> None:
        context.operations.pop(op_id, None)

    async def send_execution_result(
        self, context: ConnectionContext, op_id: str, result: ExecutionResult
    ) -> None:
        payload = {
            'data': result.data,
            'errors': [format_error(error) for error in result.errors] if result.errors else None,
        }
        await self.send_message(
            context, MessageType.GQL_DATA, op_id=op_id, payload=payload,
        )

    async def send_message(
        self, context: ConnectionContext, type: MessageType, op_id: str = None, payload: dict = None
    ) -> None:
        data = {'type': type.value}
        if op_id:
            data['id'] = op_id
        if payload:
            data['payload'] = payload
        await context.socket.send_json(data)

    async def decode(self, context: ConnectionContext, message: Message) -> OperationMessage:
        if message.get("text") is not None:
            text = message["text"]
        else:
            text = message["bytes"].decode("utf-8")

        try:
            return OperationMessage.build(json.loads(text))
        except json.decoder.JSONDecodeError:
            await context.socket.close(code=status.WS_1003_UNSUPPORTED_DATA)
            raise RuntimeError("Malformed JSON data received.")



import json
import typing

from graphql import GraphQLSchema, format_error, graphql
from starlette import status
from starlette.applications import Starlette
from starlette.background import BackgroundTasks
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, PlainTextResponse, Response
from starlette.routing import BaseRoute, Route, WebSocketRoute
from starlette.types import Receive, Scope, Send

from gql.build_schema import build_schema, build_schema_from_file
from gql.playground import PLAYGROUND_HTML
from gql.resolver import register_resolvers, default_field_resolver
from gql.utils import place_files_in_operations
from .subscribe import Subscription


class GraphQL(Starlette):
    def __init__(
        self,
        *,
        type_defs: str = None,
        schema_file: str = None,
        playground: bool = True,
        debug: bool = False,
        routes: typing.List[BaseRoute] = None,
        **kwargs,
    ):
        routes = routes or []
        if type_defs:
            self.schema = build_schema(type_defs)
        elif schema_file:
            self.schema = build_schema_from_file(schema_file)
        else:
            raise Exception('Must provide type def string or file.')
        register_resolvers(self.schema)

        routes.extend(
            [
                Route('/graphql/', ASGIApp(self.schema, playground=playground)),
                WebSocketRoute('/graphql/', Subscription(self.schema)),
            ]
        )
        super().__init__(debug=debug, routes=routes, **kwargs)


class ASGIApp:
    def __init__(self, schema: GraphQLSchema, playground: bool = True) -> None:
        self.schema = schema
        self.playground = playground

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive=receive, send=send)
        response = await self.handle_graphql(request)
        await response(scope, receive, send)

    async def handle_graphql(self, request: Request) -> Response:
        if request.method in ('GET', 'HEAD'):
            if 'text/html' in request.headers.get('Accept', ''):
                if not self.playground:
                    return PlainTextResponse('Not Found', status_code=status.HTTP_404_NOT_FOUND)
                return HTMLResponse(PLAYGROUND_HTML)

            data = request.query_params  # type: typing.Mapping[str, typing.Any]

        elif request.method == 'POST':
            content_type = request.headers.get('Content-Type', '')

            if 'application/json' in content_type:
                data = await request.json()
            elif 'application/graphql' in content_type:
                body = await request.body()
                data = {'query': body.decode()}
            elif 'query' in request.query_params:
                data = request.query_params
            elif 'multipart/form-data' in content_type:
                form = await request.form()
                try:
                    operations = json.loads(form.get('operations', '{}'))
                    files_map = json.loads(form.get('map', '{}'))
                except (TypeError, ValueError):
                    return PlainTextResponse(
                        'operations or map sent invalid JSON',
                        status_code=status.HTTP_400_BAD_REQUEST,
                    )
                data = place_files_in_operations(operations, files_map, form)
            else:
                return PlainTextResponse(
                    'Unsupported Media Type', status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
                )
        else:
            return PlainTextResponse(
                'Method Not Allowed', status_code=status.HTTP_405_METHOD_NOT_ALLOWED
            )

        try:
            query = data['query']
            variables = data.get('variables')
            operation_name = data.get('operationName')
        except KeyError:
            return PlainTextResponse(
                'No GraphQL query found in the request', status_code=status.HTTP_400_BAD_REQUEST,
            )

        background = BackgroundTasks()
        context = {'request': request, 'background': background}

        result = await graphql(
            self.schema,
            query,
            variable_values=variables,
            operation_name=operation_name,
            context_value=context,
            field_resolver=default_field_resolver,
        )
        error_data = [format_error(err) for err in result.errors] if result.errors else None
        response_data = {'data': result.data, 'errors': error_data}
        status_code = status.HTTP_400_BAD_REQUEST if result.errors else status.HTTP_200_OK

        return JSONResponse(response_data, status_code=status_code, background=background)
